// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Module: BadPixPSFCorrector
// Author: Hassan Siddiqui
//
// Description: 
//
// Point source implementation of an Encircled Energy Corrector class
//
// Modifications: See ChangeLog
// 
// $Id: BadPixPSFCorrector.cc,v 1.5 2012/03/29 13:14:00 rsaxton Exp $

/* General */
#include <cmath>

/* SAS-specific */
#include "Tracer.h"
#include "DataFilter.h"

/* Local */
#include "NumericalRange.h"
#include "BadPixPSFCorrector.h"
#include "MissingPixCorrector.h"
#include "TablePixelIterator.h"
#include "CompactTablePixelIterator.h"
#include "DetmapFilter.h"
#include "ArfDataServer.h"

/* 
   Method: BadPixPSFCorrector(string table, unsigned eefactor, 
               PsfModel * psf, unsigned int ccdid)
         : Constructor
*/
BadPixPSFCorrector::BadPixPSFCorrector(const string & table, 
  unsigned eefactor, PsfModel * psf, unsigned int ccdid):
  MissingPixCorrector(eefactor),
  _psf(psf),
  _table(table),
	_ccdid(ccdid)
{
  PROC;

  /* Create a special table iterator for the bad pixel table */
    _ctpxitr = new CompactTablePixelIterator(_table, "RAWX",
                                "RAWY","TYPE","YEXTENT");

  ENDPROC;
}
/* Destructor */
BadPixPSFCorrector::~BadPixPSFCorrector()
{
  PROC;
   //delete _pxitr;            _pxitr            =0;
   delete _detmapfilter;     _detmapfilter      =0;
   delete _ctpxitr;          _ctpxitr            =0;
  ENDPROC;
}

/* 
   Method: filterMap()
         : Removes bad pixels outside source region 
*/
void BadPixPSFCorrector::filterMap() 
{
  PROC;

  /* Set the CCD number in the CAL before filtering */
    calServer.state()->ccd()->set(_ccdid);

  /* Filter the bad pixel table so that only elements within the source 
     area remain.  Use the extension name as the local temporary filename
     NB: Cant be done in constructor because needs arfDataServer to 
     be fully set-up first */

    string localbadfile;
    if (_table.substr(0,_table.find("l")+1) == "temp_badcol") {
      localbadfile = strstr(_table.c_str(), "BADPIX");
      localbadfile = "OFFSET"+localbadfile.substr(6,2);
    } else {
      localbadfile = strstr(_table.c_str(), "BADPIX");
    }
    // Make tempfilename unique - remove for the moment
    //string unqlocbadfile = Filename::make_temp_filename(localbadfile);
    string unqlocbadfile = localbadfile;
    errstr << "BadPixPSFCorrector: filtering map "<<localbadfile
	   << Sas::message(Msg::App, Msg::Verbose);
      
    _detmapfilter = new DetmapFilter();
    _pxitr = _detmapfilter->filteredMap(_ctpxitr, 
                     //localbadfile+".ds:EVENTS", 
                     unqlocbadfile+".ds:EVENTS", 
                     arfDataServer->specAttributes(), 
                     arfDataServer->dataSubspace(), true);

  ENDPROC;
}

double BadPixPSFCorrector::eeFraction( double energy  ) const
{
  PROC;

  /* Set the CCD number in the CAL before filtering */
  calServer.state()->ccd()->set(_ccdid);

  double fracenclosed   = 0;
  //double contrib = 0;

      errstr << "BadPixPSFCorrector: Npix: "<<_pxitr->size()
	     << Sas::message(Msg::App, Msg::Verbose);
 
   /* Calculate the PSF function at this energy */
    _psf->energy(energy);

 /* Loop over each bad pixel */
/* RDS - comment out old method  */   
 /* If there are any pixels - loop over them */
//  if (_pxitr->size() > 0) {

   // FOREACH(_pxitr)
    //{
  //    PixelElement pixel = _pxitr->currentItem();
      
     /* Add contribution to PSF */
//      contrib = _psf->int_psf( pixel.xCentre(), 
        //                       pixel.yCentre(),
      //                         pixel.xWidth(),
    //                           pixel.yWidth() );
      
  //    errstr << "BadPixPSFCorrector: Pixel details  " << pixel << endl
//	     << " PSF contribution " << contrib << endl
//	     << Sas::message(Msg::App, Msg::Noisy);
      
     // fracenclosed += contrib;
   // }
 // }
 
 //// RDS - new stuff 
  double fracenclosed2 = 0.0;
  if ( _pxitr->size() > 0) 
      fracenclosed2 = _psf->enclosed_psf(_pxitr);
  
  errstr << "BadPixPSFCorrector: Fraction of psf enclosed " 
	 << fracenclosed <<" "<< fracenclosed2 <<endl
	 << Sas::message(Msg::App, Msg::Verbose); 

  RETURN fracenclosed2;
}
  
int BadPixPSFCorrector::totalPixels() const
{
  PROC;

   /* Calculate the total number of detector pixels covered by the bad 
      pixels */

     //int total = UNT(_pxitr->size() * _pxitr->wcsconvertor()->xWidth() *
      //                            _pxitr->wcsconvertor()->yWidth());

     int total = 0;

     if (_pxitr->size() > 0 ) {
       FOREACH(_pxitr)
        {
         PixelElement pixel = _pxitr->currentItem();

         total += pixel.xWidth() * pixel.yWidth();
        }  
     }

     vector<double> widths = arfDataServer->positionUtils()->detxyWidth(); 
           errstr<<"Widths: "<<widths[0]<<" "<<widths[1]
                   <<"tot for this CCD: "<<total
	     << Sas::message(Msg::App, Msg::Verbose);

  RETURN total;
}
