// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Module: CalPsfModel
// Author: Hassan Siddiqui
//
// Description:
//
// Wrapper class around CAL psf model.
//
// Modifications: See ChangeLog
//
// $Id: CalPsfModel.cc,v 1.24 2014/03/18 11:05:36 rsaxton Exp $

/* SAS */
#include "errstr.h"
#include "Tracer.h"
#include "CalServer.h"
#include "Region.h"

// Local classes
#include "ArfDataServer.h"
#include "CalPsfModel.h"
#include "PositionUtils.h"

/* Constructor */
CalPsfModel::CalPsfModel( double energy,
			  double detxcentre,
			  double detycentre,
                          AccuracyLevel mode,
			  const PositionUtils * const putils):
  _putils(putils),
  _mode(mode),
  _offset(0,0,0),
  _t(0,0),
  _boxShape(ShapeBasics::CIRCLE),
  _fastmode_factor(1.0)
{
  PROC;

  errstr
    << "Constructing CalPsfModel with energy=" << energy
    << " detxcentre=" << detxcentre
    << " detycentre=" << detycentre
    << " mode=" << _mode
    << Sas::message(Msg::App, Msg::Verbose);


  _offset = _putils->detxy005ToCamCoord2( detxcentre,
					  detycentre );

  _t = _putils->camCoord2ToTelCoord(_offset);

  calServer.state()->accuracyLevel()->set(_mode);

  PsfDataServer * psfs = 0; psfs = calServer.getAtom(psfs);

  /* Use a nasty trick to convert a reference to a pointer. */
  errstr
    << "Calling PsfDataServer for a Psf at TELCOORD position ["
    << setprecision(10) << _t.theta()
    << ","
    << setprecision(10) << _t.phi()
    << "]"
    << Sas::message(Msg::App, Msg::Verbose);

    Psf & p = psfs->psf(energy, _t.theta(), _t.phi());

    _psf = &p;

  /* Find the actual mode used in case it got changed */
  _mode = calServer.state()->accuracyLevel()->get();

  /* Reduce bin size */
  if (_psf->binSizeChangeable())
      {

//	const unsigned OVERSAMPLE_FACTOR = 4;
	/* AI 2009-07-10
	 * cal/src/PsfDataServer.h -> There was a typedef named BinSize and it was
	 * declared a variable with the same name. This throwed an error on gcc4.
	 * The variable name was renamed to "ValidBinSize"
	 */
      	Psf::ValidBinSize bz = _psf->binSize();
                // CAREFUL - REMOVED SOMETHING HERE !!!!!!!!!!!!!!!!!!!!!
//	Psf::ValidBinSize bz_out(bz.binX()/OVERSAMPLE_FACTOR, bz.binY()/OVERSAMPLE_FACTOR);
        Psf::ValidBinSize bz_out = bz;
                // CAREFUL - REMOVED SOMETHING HERE !!!!!!!!!!!!!!!!!!!!!
	errstr
	  << " Changing psf binsize from ["
	  << bz.binX() << "," << bz.binY()
	  << "] to ["
	  << bz_out.binX() << "," << bz_out.binY()
	  << "]"
	  << Sas::message(Msg::App, Msg::Noisy);
	_psf->binSize(bz_out);
  }


  ENDPROC;
};

CalPsfModel::~CalPsfModel()
{
  PROC;
  errstr
    << "CalPsfModel destroyed."
    << Sas::message(Msg::App, Msg::Verbose);

  ENDPROC;
}

/* Return PSF function at given position */
double CalPsfModel::psf( double detx,
			 double dety) const
{
  PROC;
  CamCoords c = _putils->detxy005ToCamCoord2( detx,
					      dety );

  double out = (*_psf)(c.x(), c.y());

  errstr
    << "CalPsfModel::psf: for detx=" << detx
    << " dety=" << dety
    << " get a psf of " << out
    << Sas::message(Msg::App, Msg::Noisy);

  RETURN out;
}

/* Return the encircled energy from a selection of pixels.
   Input is in detector coords */
double CalPsfModel::enclosed_psf( PixelIterator * pxitr ) const
{

  double total = 0.0;  // Initialise total psf fraction

 // Set some constants
  double sec2mm = 0.03636;
  double binsize = 20.0;

 // Get the PSF image
  Image<float> psfmap = _psf->image(SpatialWindow<double>(300.0*sec2mm, 
                                                           300.0*sec2mm));

  // DEBUG
  // Get and siaplay total of PSF image
//double deltot = 0.0;
//for (unsigned ix=0; ix<psfmap.xExtent(); ++ix) 
  //   for (unsigned iy=0; iy<psfmap.yExtent(); ++iy) 
    //          deltot += psfmap[ix][iy];
//cout<<"PSF integration: "<<deltot<<endl;
  // DEBUG

  int xhalf = INT (psfmap.xExtent() / 2) ;
  int yhalf = INT (psfmap.yExtent() / 2);
  int xmax = psfmap.xExtent()-1, ymax = psfmap.yExtent()-1;
  double pbit=0.0;
  double norm=0.0;

 // Convert PSF centre back to detector coordinates
  CoordPair detoff = _putils->CamCoord2Todetxy005(_offset);

 // The round function used below has the property of rounding -0.5 to -1 and
 // +0.5 to 1. This can lead to the central pixel of the PSF (300,300)
 // being undersampled and hence the PSF underestimated. To avoid this we
 // make sure that the offset positions used here are non-integer.
  double dxoff=detoff.x(), dyoff=detoff.y();
  if (detoff.x() == double(int(detoff.x()))) dxoff=detoff.x()+0.001;
  if (detoff.y() == double(int(detoff.y()))) dyoff=detoff.y()+0.001;

  FOREACH(pxitr)
    {
      PixelElement pixel = pxitr->currentItem();

     // Convert from input detector coords into PSF-centred mm frame
      //int nx= round((pixel.xCentre()-detoff.x()) / binsize) + xhalf;
      int nx= round((double(pixel.xCentre())-dxoff) / 
                               double(binsize)) + xhalf;
      //int ny= round((pixel.yCentre()-detoff.y()) / binsize) + yhalf;
      int ny= round((double(pixel.yCentre())-dyoff) / 
                               double(binsize)) + yhalf;

     /* Add the PSF contribution at the position of this pixel 
        normalised by the width of the pixel (CAL returns 1"x1" 
        value). NB: PSF is normalised to 1.0 for a 300" extraction 
        radius but is non-zero outside so must explicitely exclude 
        larger X and Y values */

      pbit = 0.0;  // DEBUG REMOVE
      if (nx >= 0 && nx <=xmax && ny >= 0 && ny <=ymax) {

       // Calculate the normalisation factor in arcsec squared
        //double cwidx = pixel.xWidth() / 20.0, cwidy = pixel.yWidth() / 20.0;
         norm = pixel.xWidth() * pixel.yWidth() / 400.0;

         pbit = psfmap[nx][ny]*norm; 
         total += pbit;
      }

      //cout<<"Cam Coord: "<<pixel.xCentre()<<" "<<pixel.yCentre()<<" "<<nx<<" "<<ny<<" "<<norm<<" "<<pbit<<endl;
    }

  errstr
    << "CalPsfModel::enclosed psf=" << total
    //<< Sas::message(Msg::App, Msg::Noisy);
    << Sas::message(Msg::App, Msg::Verbose);  // RDS PUT BACK !

    /* Finally correct the total if a subset of the extraction 
       region in burst mode has been used */
    if (_fastmode_factor > 0.0 && _fastmode_factor < 1.0)
                      total = total * _fastmode_factor;

  RETURN total;
}

/* Calculate integrated PSF over a rectangular area. */
double CalPsfModel::int_psf( double detxboxcentre,
			     double detyboxcentre,
			     double detxboxsize,
			     double detyboxsize ) const
{
  PROC;

  /* Calculate invariate properties if this is the first time through */
  /* Know that camcoords2/det have same origin (0,0).
     So can use same method for scales as well as positions */
  CamCoords c2=_putils->detxy005ToCamCoord2( detxboxsize,
                                               detyboxsize );
  CamCoords bit=_putils->detxy005ToCamCoord2(1,1);
  double halfx = c2.x() / 2.0;
  double pifactor = c2.x() / (2.0 * M_PI);

  CamCoords c1 = _putils->detxy005ToCamCoord2( detxboxcentre,
					       detyboxcentre );

  double xdiff=c1.x()-_offset.x(), ydiff=c1.y()-_offset.y();


  double out=0.0;


  /* For HIGH and MEDIUM mode use the enboxedEnergy code */
  if (_mode != EXTENDED && _mode != ELLBETA) {
  //if (_mode != EXTENDED)  {

  /* This call is needed for the HIGH mode but shouldn't be called
     for ELLBETA */
     _psf->atOrigin(xdiff,ydiff);   

     Psf::PsfWindow pw(c2.x()/2., c2.y()/2);

     out = _psf->enboxedEnergy(pw);

     errstr
     << "CalPsfModel::int_psf: for "
     << " energy=" << _psf->E()
     << " theta=" << _t.theta()
     << " phi=" << _t.phi()
     << endl
     << " detxboxcentre=" << detxboxcentre
     << " detyboxcentre=" << detyboxcentre
     << " detxboxsize=" << detxboxsize
     << " detyboxsize=" << detyboxsize
     << endl
     << " derive a PsfWindow of "
     << " halfX=" << pw.halfX()
     << " halfY=" << pw.halfY()
     << endl
     << " and get an enclosed energy " << out
     << Sas::message(Msg::App, Msg::Noisy);

   } else {   // circular, elliptical  or annular source box calculation

/* The above doesn't work very well for circular or annular source
   boxes - so use the following technique
      Algorithm: Find the total PSF within the outer radius of
      the box, subtract the total PSF from within the inner radius
      and normalise for the box size over the annulus area
*/

   /* Calculate the mean radius of this box */
      double boxRadius = 0.0;

      if (_mode == EXTENDED)
         boxRadius = sqrt( xdiff*xdiff + ydiff*ydiff );
      else if (_mode == ELLBETA)
         boxRadius = getMajorAxis(xdiff, ydiff);

   /* This technique doesn't work for pixels right on the source
      centre so make sure radius is bigger than half a box */
      if (boxRadius < halfx + bit.x() ) {
           boxRadius = halfx + bit.x();
      }

   /* Find the PSF contribution in an annulus about this radius (in arcsec)  */
      double PsfCont = _psf->encircledEnergy(boxRadius + halfx) -
                       _psf->encircledEnergy(boxRadius - halfx);

   /* Normalise by area to get the PSF fraction in the box */
     // FOR circle
      out = PsfCont * pifactor / boxRadius;

      //cout<<"PSFBIT: "<<PsfCont<<" "<<pifactor<<" "<<boxRadius<<" "<<halfx
        //  <<" "<<_psf->encircledEnergy(boxRadius + halfx)<<" "<<
          //  _psf->encircledEnergy(boxRadius - halfx)<<endl;

     // FOR ellipse, correct for the ellipticity
      if (_mode == ELLBETA)
             out = out / sqrt(1.0 - _psf->ellipticity() * _psf->ellipticity());

   }

  RETURN out;
}

/* Set energy */
void CalPsfModel::energy(double energy )
{
  PROC;

  _psf->setParameters(energy, _t.theta(), _t.phi());

  ENDPROC;
}

string CalPsfModel::mode() const
{
  PROC;

  RETURN "Dummy";
}

void CalPsfModel::setRadius( vector<double> radii )
{

  PROC;
    for (unsigned i=0; i<radii.size(); i++) {

      errstr
        << "Setting radius "<<i<<" = " <<radii[i]
        << Sas::message(Msg::App, Msg::Verbose);

       _radii.push_back(radii[i]);
    }
      errstr
        << "Finished for i="<<radii.size()
        << Sas::message(Msg::App, Msg::Verbose);
  ENDPROC;
}

void CalPsfModel::setNgridEls(unsigned int /* NGridEls */)
{
  PROC;
  /* Do nothing */
  ENDPROC;
}


double CalPsfModel::rad_psf() const
{
  PROC;

      double ee=0;

      errstr
        << "Getting Encircled energy (energy): "<<_psf->E()
        << " radius "<<_radii[0]
        << Sas::message(Msg::App, Msg::Verbose);

  /* Calculation method depends on source box shape */
      if ( _boxShape == ShapeBasics::CIRCLE ||
           _boxShape == ShapeBasics::ANNULUS ||
           _boxShape == ShapeBasics::ELLIPSE ||
           _boxShape == ShapeBasics::ELLIPTANNULUS) {

	ee = _psf->encircledEnergy(_radii[0]);

	/* Subtract the encircled energy from within the inner radius if
	   this is an annulus */

	if (_radii.size() == 2) {
	  ee = _psf->encircledEnergy(_radii[1]) - ee;
	}

//	cout<<"EE: "<<_psf->encircledEnergy(_radii[0])<<" "<<_psf->encircledEnergy(_radii[1])<<" "<<_radii[0]<<" "<<_radii[1]<<endl;
      } else {   // BOX


         double ysmall=0.0, ybig=0.0;
         //double minor=0.0, major=0.0;

  /* For PN Timing/Burst mode need to treat the length in two pieces
     depending on the exact position of the source */
         Xmm::Mode windowMode = calServer.state()->mode()->get();
         if (windowMode == Xmm::FAST_TIMING ||
                     windowMode == Xmm::FAST_BURST)
         {
            ModeParameters *modeParam = 0;
                              modeParam = calServer.getAtom(modeParam);

            MiscDataServer * mds = 0; mds = calServer.getAtom(mds);
            double plateY = mds->get("PLATE_SCALE_Y");

          /* Set the centre of the source in RAW-Y */
            int PNFASTMODE_RAWYPOS = _SRCPOS;

          /* Calc the distance from nominal source centre to chip border
             in CamCoords (mm) */
            //ysmall = (modeParam->windowY0()+modeParam->windowDY()-1
            ysmall = (modeParam->windowY0()+modeParam->windowDY()
                      - PNFASTMODE_RAWYPOS); // In RAW pixels
            ysmall = _putils->detxy005ToCamCoord2(INT(ysmall*plateY*20.0),1).x();
            //cout<<"MP: "<<modeParam->windowY0()<<" "<<modeParam->windowDY()
              //   <<" "<<plateY<<endl;

          // Calc the size of what is left of the Y box
            ybig = _radii[1] * 2 - ysmall;

            errstr<<"YSMALL: "<<ysmall<<" "<<ybig<<" SRCPOS used: "
                  <<PNFASTMODE_RAWYPOS
                  << Sas::message(Msg::App, Msg::Noisy);

         } else {
            //ybit->push_back(_radii[1]);
            ysmall=_radii[1];
            ybig=_radii[1];
         }

         double xbit= _radii[0]; // The X radius - half-width

  /* Let us try getting the image and summing the bit of the image
     that we want. Note that this assumes that the image returned is
     normalised to 1.0 which is not always the case. */

     // Set the bin size for the window to be 1". This is necessary
     // because the image is always returned integrated with 1" x 1" pixels.
         double sec2mm = 0.03636;
         int xcent = 300;
         int ycent=410;

         if (_psf->binSizeChangeable())
             _psf->binSize(Psf::ValidBinSize(sec2mm,sec2mm));  // Temp hardcode

     // Set up PSF window - of 820" x 600"  in 1" x 1" pixels to cover the
     // extent of the PN and MOS chips
	 //Image<float> psfmap = _psf->image(Psf::PsfWindow(300.0*sec2mm, 410.0*sec2mm);
	 Image<float> psfmap = _psf->image(SpatialWindow<double>(300.0*sec2mm, 410.0*sec2mm));
	 //Image<float> psfmap = _psf->image(SpatialWindow<double>(300.0*sec2mm, 300.0*sec2mm));

     // Now sum over the pixels that we want

  // Convert X, Y ranges into arcsecs on the grid
         int xstart=max(1,xcent - INT(xbit / sec2mm));
         int xstop=min(xcent*2,xcent + INT(xbit / sec2mm) -1);  // One too many ?
         int ystart=max(1,ycent - INT(ysmall / sec2mm));
         int ystop=min(ycent*2,ycent + INT(ybig / sec2mm)-1);  // One too many ?

         for (int nx=xstart; nx<=xstop; ++nx)
           for (int ny=ystart; ny<=ystop; ++ny)
               ee +=psfmap[nx][ny];

         // DEBUG !!!!!
         //psfmap = _psf->image(SpatialWindow<double>(2*sec2mm, 2*sec2mm));

    //     cout<<" im(0): "<<psfmap[0][0]<<" "<<psfmap[0][1]<<" "<<psfmap[0][2]<<" "<<psfmap[0][3]<<" "<<psfmap[0][4]<<endl;
      //   cout<<" im(1): "<<psfmap[1][0]<<" "<<psfmap[1][1]<<" "<<psfmap[1][2]<<" "<<psfmap[1][3]<<" "<<psfmap[1][4]<<endl;
   //      cout<<" im(2): "<<psfmap[2][0]<<" "<<psfmap[2][1]<<" "<<psfmap[2][2]<<" "<<psfmap[2][3]<<" "<<psfmap[2][4]<<endl;
     //    cout<<" im(3): "<<psfmap[3][0]<<" "<<psfmap[3][1]<<" "<<psfmap[3][2]<<" "<<psfmap[3][3]<<" "<<psfmap[3][4]<<endl;
    //     cout<<" im(4): "<<psfmap[4][0]<<" "<<psfmap[4][1]<<" "<<psfmap[4][2]<<" "<<psfmap[4][3]<<" "<<psfmap[4][4]<<endl;
         
     /* We have a normalisation issue here. The image pixels are given
        at 1" resolution but our area can be any size. Multiply the ee
        by the difference in areas but this isn't right because the PSF
        is dropping off fast - at least in the middle. Don't make
        this "correction" */
      //   double integrated_size=abs( (xstop - xstart + 1) * 
        //                                 (ystop - ystart + 1) );
      //   double real_size = 2.0 * xbit * 
        //                     (ystop - ycent + 1 + ysmall / sec2mm) / sec2mm;
      //   ee = min(1.0, ee * real_size / integrated_size);

         //cout<<"ST: "<<xstart<<" "<<xstop<<" "<<ystart<<" "<<ystop
           //  <<" XSM: "<<xbit<<" "<<ybig<<" "<<ysmall
             // <<" "<<integrated_size<<" "<<real_size<<endl;
 
        // cout<<" im(297): "<<psfmap[297][408]<<" "<<psfmap[297][409]
               //            <<psfmap[297][410]<<endl;
      }

    /* Finally correct this value if a subset of the extraction region
       in burst mode has been used */
     if (_fastmode_factor > 0.0 && _fastmode_factor < 1.0)
                      ee = ee * _fastmode_factor;

  RETURN ee;
}

/* Method - setBoxShape() - sets the input shape */
void CalPsfModel::setBoxShape(ShapeBasics::ShapeType shape)
{
  PROC;

     _boxShape = shape;

  ENDPROC;
}

/* Method - setFastModeFactor - sets a factor to scale the ee correction */
void CalPsfModel::setFastModeFactor(double factor)
{
   _fastmode_factor = factor;
}

/* Method - setFastModeFactor - sets a factor to scale the ee correction */
void CalPsfModel::setSRCPOS(int SRCPOS)
{
   _SRCPOS = SRCPOS;
}

/* Method - setOrientation - sets the orientation in radians */
void CalPsfModel::setOrientation(double orient)
{
   _orient = orient * DEG_TO_RAD;
}

/* Method - getMajorAxis - calculates the major axis length of the ellipse
                           which includes the point at a distance
			   xdiff, ydiff from the centre */
double CalPsfModel::getMajorAxis(double xdiff, double ydiff) const
{
  PROC;

   double a=0.0;

  /* Calculate the polar angle of this point around the ellipse */
   //double theta = 0.0;
   //if ( (x - _offset.x()) != 0.0)
   double theta = atan2(xdiff,ydiff) - _orient;
   //else
     // theta = M_PI / 2.0 - _orient;

  /* Calculate the radius at this point */
   //double r = abs(x - _offset.x()) / cos(theta + _orient);
   double r = sqrt(ydiff * ydiff + xdiff * xdiff);

  /* Calculate the semi-minor axis length */
   double pe2 = _psf->ellipticity() * _psf->ellipticity();
   double ct = cos(theta);

   double b = r * sqrt(1.0 - pe2 * ct * ct);

  /* Calculate the semi-major axis length */
   a = b / sqrt(1.0 - pe2);

   //cout<<" Y: "<<ydiff<<" "<<xdiff<<" "<<a<<endl;
   //cout<<"THETA: "<<theta<<" "<<r<<" "<<b<<" "<<a<<" "<<_psf->ellipticity()
       //<<_orient<<endl;

   RETURN a;
}
